"""Data tools for FINAM."""

from .core import (
    assert_type,
    check,
    full,
    full_like,
    has_time_axis,
    prepare,
    strip_time,
    to_datetime,
)
from .info import Info
from .mask import (
    Mask,
    check_data_covers_domain,
    filled,
    from_compressed,
    has_masked_values,
    is_masked_array,
    is_sub_mask,
    mask_specified,
    masks_compatible,
    masks_equal,
    to_compressed,
    to_masked,
)
from .units import (
    UNITS,
    check_quantified,
    clear_units_cache,
    compatible_units,
    equivalent_units,
    get_dimensionality,
    get_magnitude,
    get_units,
    is_quantified,
    quantify,
    to_units,
)

__all__ = [
    "UNITS",
    "Info",
    "Mask",
    "assert_type",
    "check",
    "check_data_covers_domain",
    "check_quantified",
    "clear_units_cache",
    "compatible_units",
    "equivalent_units",
    "filled",
    "from_compressed",
    "full",
    "full_like",
    "get_dimensionality",
    "get_magnitude",
    "get_units",
    "has_masked_values",
    "has_time_axis",
    "is_masked_array",
    "is_quantified",
    "is_sub_mask",
    "mask_specified",
    "masks_compatible",
    "masks_equal",
    "prepare",
    "quantify",
    "strip_time",
    "to_compressed",
    "to_datetime",
    "to_masked",
    "to_units",
]
